
PROMPT_TOO_LONG_ERROR = "It's not possible to process your request because the prompt is too long. Please open a new conversation and try again with a shorter prompt."
MODEL_NOT_FOUND_ERROR = "There was an error processing your request because the LLM was not found. Contact the administrator."
REQUEST_ERROR = "There was an error processing your request. Please try again or contact the administrator."

class PromptTooLongException(Exception):
    def __init__(self, message=PROMPT_TOO_LONG_ERROR):
        self.message = message
        super().__init__(self.message)

class ModelNotFoundException(Exception):
    def __init__(self, message=MODEL_NOT_FOUND_ERROR):
        self.message = message
        super().__init__(self.message)

class RequestException(Exception):
    def __init__(self, message=REQUEST_ERROR):
        self.message = message
        super().__init__(self.message)

    


from logging import getLogger
from typing import Any, Dict, List

import numpy as np
from transformers import pipeline

logger = getLogger(__name__)

models_dict = {
    'en': 'MoritzLaurer/DeBERTa-v3-base-mnli-fever-anli', # 'MoritzLaurer/DeBERTa-v3-large-mnli-fever-anli-ling-wanli' for little bit better performance
    'es': 'MoritzLaurer/mDeBERTa-v3-base-mnli-xnli',
    'fr': 'MoritzLaurer/mDeBERTa-v3-base-mnli-xnli',
    'multi': 'MoritzLaurer/mDeBERTa-v3-base-mnli-xnli', # 'MoritzLaurer/multilingual-MiniLMv2-L6-mnli-xnli' for a little bit worse performance but faster
}
    

class ReferenceChecker:
    def __init__(self, lang: str = "en", device: str = 'cuda'):
        """
        Class to check if we need to retrieve new contexts. We use a zero-shot classifier to check if the message is a some kind of question or instruction, if not we don't need to retrieve new contexts.
        """
        lang = lang if lang in models_dict else 'multi' # default to multilingual model if language not supported

        self.candidate_labels = ['greeting', 'farewell', 'personal check-in', 'small talk', "acknowledgments", 'question', 'instruction']
        self.true_labels = ['question', 'instruction']
        self.reference_checker = pipeline('zero-shot-classification', model=models_dict[lang], device=device)
        logger.info(f"Loaded reference checker for language {lang}")


    def check_references(self, message: str) -> bool:
        """
        Check if the message is a question or instruction. If so, we need to retrieve new contexts.
        """
        output = self.reference_checker(message, self.candidate_labels, multi_label=False)
        label = output['labels'][0]
        if label in self.true_labels: # if label is question or instruction, then retrieve
            return True
        
        # For doubful cases, we retrieve new contexts, 
        # e.g. if the label is not question or instruction but the score of one of those is close to the highest score
        highest_score = output['scores'][0]
        for true_label in self.true_labels:
            if (highest_score - output['scores'][output['labels'].index(true_label)]) < 0.1:
                return True

        return False # if not question or instruction, then don't retrieve
    

def clean_relevant_references(sources: List[Dict[str, Any]], min_difference=0.09, min_score=0.3) -> List[Dict[str, Any]]:
    """
    Find the relevant sources from a list of sources. We use the similarity scores to find the relevant sources. We calculate the standard deviation of the gaps between consecutive sources and return all sources up to the first significant gap.
    Parameters
    ----------
    sources: List[Dict[str, Any]]
        List of sources to find the relevant sources from.
    min_difference: float
        Minimum difference between consecutive sources to be considered a significant gap.
    min_score: float
        Minimum score for a source to be considered relevant.
    Returns
    -------
    List[Dict[str, Any]]
        List of relevant sources.
    """

    # Calculate gaps between consecutive sources
    # print the similarity scores
    gaps = [sources[i]['score'] - sources[i + 1]['score'] for i in range(len(sources) - 1)]

    # Compute standard deviation of the gaps if there are enough gaps
    if len(gaps) > 1:
        std_dev = np.std(gaps)
    else:
        # If there's only one gap or none, return all sources as relevant
        return sources

    # Identify the first significant gap
    for i, gap in enumerate(gaps):
        if gap > std_dev and gap > min_difference:  # A significant gap
            return sources[:i + 1]
        
    final_sources = []
    for source in sources:
        if source['score'] > min_score:
            final_sources.append(source)

    # If no significant gap found, return all sources
    return final_sources


import os
from typing import Dict, List

from openai import AsyncOpenAI, OpenAI

from chat_rag.llms import RAGLLM


class OpenAIChatModel(RAGLLM):
    def __init__(
        self,
        llm_name: str,
        **kwargs,
    ):
        self.client = OpenAI(api_key=os.environ["OPENAI_API_KEY"])
        self.llm_name = llm_name

    def format_prompt(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        system_prefix: str,
        n_contexts_to_use: int = 3,
        lang: str = "en",
        **kwargs,
    ) -> List[Dict[str, str]]:
        """
        Formats the prompt to be used by the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : list
            The context to use.
        system_prefix : str
            The prefix to indicate instructions for the LLM.
        system_tag : str
            The tag to indicate the start of the system prefix for the LLM.
        system_end : str
            The tag to indicate the end of the system prefix for the LLM.
        user_tag : str
            The tag to indicate the start of the user input.
        user_end : str
            The tag to indicate the end of the user input.
        assistant_tag : str
            The tag to indicate the start of the assistant output.
        assistant_end : str
            The tag to indicate the end of the assistant output.
            The tag to indicate the end of the role (system role, user role, assistant role).
        n_contexts_to_use : int, optional
            The number of contexts to use, by default 3
        lang : str, optional
            The language of the prompt, by default 'en'
        """
        system_prompt = self.format_system_prompt(
            contexts=contexts,
            system_prefix=system_prefix,
            n_contexts_to_use=n_contexts_to_use,
            lang=lang,
        )

        final_messages = [{'role': 'system', 'content': system_prompt}] + messages

        return final_messages

    def generate(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ) -> str:
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        messages = self.format_prompt(
            messages=messages,
            contexts=contexts,
            **prompt_structure_dict,
            lang=lang,
        )

        response = self.client.chat.completions.create(model=self.llm_name,
        messages=messages,
        max_tokens=generation_config_dict["max_new_tokens"],
        temperature=generation_config_dict["temperature"],
        top_p=generation_config_dict["top_p"],
        presence_penalty=generation_config_dict["repetition_penalty"],
        seed=generation_config_dict["seed"],
        n=1,
        stream=False)
        return response.choices[0].message.content

    def stream(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ):
        """
        Generate text from a prompt using the model in streaming mode.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        messages = self.format_prompt(
            messages=messages,
            contexts=contexts,
            **prompt_structure_dict,
            lang=lang,
        )

        print(messages)

        response = self.client.chat.completions.create(model=self.llm_name,
        messages=messages,
        max_tokens=generation_config_dict["max_new_tokens"],
        temperature=generation_config_dict["temperature"],
        top_p=generation_config_dict["top_p"],
        presence_penalty=generation_config_dict["repetition_penalty"],
        seed=generation_config_dict["seed"],
        n=1,
        stream=True)
        for chunk in response:
            if chunk.choices[0].finish_reason == "stop":
                return
            if chunk.choices[0].delta.content is not None:
                yield chunk.choices[0].delta.content # return the delta text message


class AsyncOpenAIChatModel(RAGLLM):
    def __init__(
        self,
        llm_name: str,
        **kwargs,
    ):
        self.client = AsyncOpenAI(api_key=os.environ["OPENAI_API_KEY"])
        self.llm_name = llm_name

    def format_prompt(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        system_prefix: str,
        n_contexts_to_use: int = 3,
        lang: str = "en",
        **kwargs,
    ) -> List[Dict[str, str]]:
        """
        Formats the prompt to be used by the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : list
            The context to use.
        system_prefix : str
            The prefix to indicate instructions for the LLM.
        system_tag : str
            The tag to indicate the start of the system prefix for the LLM.
        system_end : str
            The tag to indicate the end of the system prefix for the LLM.
        user_tag : str
            The tag to indicate the start of the user input.
        user_end : str
            The tag to indicate the end of the user input.
        assistant_tag : str
            The tag to indicate the start of the assistant output.
        assistant_end : str
            The tag to indicate the end of the assistant output.
            The tag to indicate the end of the role (system role, user role, assistant role).
        n_contexts_to_use : int, optional
            The number of contexts to use, by default 3
        lang : str, optional
            The language of the prompt, by default 'en'
        """
        system_prompt = self.format_system_prompt(
            contexts=contexts,
            system_prefix=system_prefix,
            n_contexts_to_use=n_contexts_to_use,
            lang=lang,
        )

        final_messages = [{'role': 'system', 'content': system_prompt}] + messages

        return final_messages

    async def generate(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ) -> str:
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        messages = self.format_prompt(
            messages=messages,
            contexts=contexts,
            **prompt_structure_dict,
            lang=lang,
        )

        response = await self.client.chat.completions.create(model=self.llm_name,
        messages=messages,
        max_tokens=generation_config_dict["max_new_tokens"],
        temperature=generation_config_dict["temperature"],
        top_p=generation_config_dict["top_p"],
        presence_penalty=generation_config_dict["repetition_penalty"],
        seed=generation_config_dict["seed"],
        n=1,
        stream=False)
        return response.choices[0].message.content
    
    async def stream(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ):
        """
        Generate text from a prompt using the model in streaming mode.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        messages = self.format_prompt(
            messages=messages,
            contexts=contexts,
            **prompt_structure_dict,
            lang=lang,
        )

        print(messages)

        response = await self.client.chat.completions.create(model=self.llm_name,
        messages=messages,
        max_tokens=generation_config_dict["max_new_tokens"],
        temperature=generation_config_dict["temperature"],
        top_p=generation_config_dict["top_p"],
        presence_penalty=generation_config_dict["repetition_penalty"],
        seed=generation_config_dict["seed"],
        n=1,
        stream=True)
        async for chunk in response:
            if chunk.choices[0].finish_reason == "stop":
                return
            if chunk.choices[0].delta.content is not None:
                yield chunk.choices[0].delta.content


from typing import List
from logging import getLogger

from chat_rag.inf_retrieval.cross_encoder import ReRanker


logger = getLogger(__name__)


class ReRankRetriever:
    """
    Class for reranking retrieved contexts on top of the retrieval model.
    """
    def __init__(self, retriever, lang, device):
        self.retriever = retriever
        self.reranker = ReRanker(lang=lang, device=device)

    def retrieve(
        self,
        queries: List[str],
        top_k: int = 5,
    ):
        contexts_retrieved = self.retriever.retrieve(queries, top_k=top_k)  # retrieve contexts
        contexts_ranked = []
        for query, contexts in zip(queries, contexts_retrieved):
            query_contexts = self.reranker(query, contexts)  # rerank and filter contexts
            contexts_ranked.append(query_contexts)
        return contexts_ranked



import json
import os
from typing import Dict, Iterable, List, AsyncIterable

import aiohttp
import requests
from openai import OpenAI, AsyncOpenAI

from chat_rag.llms import RAGLLM
from chat_rag.exceptions import ModelNotFoundException, RequestException, PromptTooLongException

import logging

logger = logging.getLogger(__name__)


class VLLMModel(RAGLLM):
    """
    A client that sends requests to the VLLM server.
    """

    def __init__(
        self, llm_name: str, base_url: str = None, use_openai_api: bool = True, **kwargs
    ):
        super().__init__(llm_name=llm_name, **kwargs)
        if base_url is None:
            self.endpoint_url = os.environ["VLLM_ENDPOINT_URL"]
        else:
            self.endpoint_url = base_url

        self.llm_name = llm_name

        # I could use the already OpenAI implementation, but I prefer to implement the OpenAI API here also
        # because I need to do checks on the prompt length before sending it to the API
        # and I cannot do that on the OpenAI implementation
        if use_openai_api:
            self.client = OpenAI(
                base_url=self.endpoint_url
            )  # for VLLM OpenAI compatible API
        self.use_openai_api = use_openai_api
        print(f"Using vLLM OpenAI compatible API server: {self.use_openai_api}")

    def _format_prompt_openai(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        system_prefix: str,
        lang: str = "en",
        **kwargs,
    ) -> List[Dict[str, str]]:
        """
        Formats the prompt to be used by the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : list
            The context to use.
        lang : str, optional
            The language of the prompt, by default 'en'
        """
        system_prompt = self.format_system_prompt(
            contexts=contexts,
            system_prefix=system_prefix,
            lang=lang,
        )

        final_messages = [{"role": "system", "content": system_prompt}] + messages

        return final_messages

    def _generate_vllm(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> List[str]:
        """
        Generate text from a prompt using the vLLM API server.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        stop_words : List[str]
            The stop words to use to stop generation.
        Returns
        -------
        str
            The generated text.
        """

        prompt, _ = self.format_prompt(
            messages, contexts, **prompt_structure_dict, lang=lang
        )

        pload = {
            "prompt": prompt,
            "n": 1,
            "use_beam_search": False,
            "top_k": generation_config_dict["top_k"],
            "top_p": generation_config_dict["top_p"],
            "temperature": generation_config_dict["temperature"],
            "max_tokens": generation_config_dict["max_new_tokens"],
            "frequency_penalty": generation_config_dict["repetition_penalty"],
            "stop": stop_words,
            "stream": False,
        }

        response = requests.post(self.endpoint_url, json=pload, stream=False)

        if response.status_code != 200:
            logger.error(f"Error with the request to the vLLM server: {response.content}")
            raise RequestException()


        data = json.loads(response.content)

        # return the difference between the prompt and the output
        output = data["text"][0][len(prompt) :]

        if not output: # if there is an error vllm returns an empty string
            logger.error(f"Error with the request to the vLLM server.")
            raise RequestException()

        return output

    def _generate_openai(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ) -> str:
        """
        Generate text from a prompt using the vLLM OpenAI API.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        # Check how many messages and contexts we can fit in the model context length constraints
        _, (n_contexts, n_messages_to_keep) = self.format_prompt(
            messages, contexts, **prompt_structure_dict, lang=lang
        )

        messages = self._format_prompt_openai(
            messages=messages[:n_messages_to_keep],  # keep only the last n_messages_to_keep
            contexts=contexts[:n_contexts],  # keep only the last n_contexts
            **prompt_structure_dict,
            lang=lang,
        )

        try:
            print(f'LLM name: {self.llm_name}')
            response = self.client.chat.completions.create(
                model=self.llm_name,
                messages=messages,
                max_tokens=generation_config_dict["max_new_tokens"],
                temperature=generation_config_dict["temperature"],
                top_p=generation_config_dict["top_p"],
                presence_penalty=generation_config_dict["repetition_penalty"],
                seed=generation_config_dict["seed"],
                n=1,
                stream=False,
            )
            return response.choices[0].message.content
        except Exception as e:
            return return_openai_error(e)

    def generate(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> str | List[str] | None:
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        stop_words : List[str]
            The stop words to use to stop generation.
        Returns
        -------
        str
            The generated text.
        """

        if self.use_openai_api:
            return self._generate_openai(
                messages=messages,
                contexts=contexts,
                prompt_structure_dict=prompt_structure_dict,
                generation_config_dict=generation_config_dict,
                lang=lang,
                **kwargs,
            )
        else:
            return self._generate_vllm(
                messages=messages,
                contexts=contexts,
                prompt_structure_dict=prompt_structure_dict,
                generation_config_dict=generation_config_dict,
                lang=lang,
                stop_words=stop_words,
                **kwargs,
            )

    def _stream_vllm(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> Iterable[str]:
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        stop_words : List[str]
            The stop words to use to stop generation.
        Returns
        -------
        str
            The generated text.
        """

        prompt, _ = self.format_prompt(
            messages, contexts, **prompt_structure_dict, lang=lang
        )

        pload = {
            "prompt": prompt,
            "n": 1,
            "use_beam_search": False,
            "top_k": generation_config_dict["top_k"],
            "top_p": generation_config_dict["top_p"],
            "temperature": generation_config_dict["temperature"],
            "max_tokens": generation_config_dict["max_new_tokens"],
            "frequency_penalty": generation_config_dict["repetition_penalty"],
            "stop": stop_words,
            "stream": True,
        }

        response = requests.post(self.endpoint_url, json=pload, stream=True)

        prev_output = pload["prompt"]
        for n_token, chunk in enumerate(response.iter_lines(
            chunk_size=8192, decode_unicode=False, delimiter=b"\0"
        )):
            if n_token == 1 and not output: # if there is an error vllm returns an empty string as the second chunk (the first one is the prompt)
                raise RequestException()
            if chunk:
                data = json.loads(chunk.decode("utf-8"))
                if 'detail' in data:
                    logger.error(f"Error with the request to the vLLM server: {data['detail']}")
                    raise RequestException()
                output = data["text"]
                # yield the difference between the previous output and the current output
                output = output[0][len(prev_output) :]
                prev_output += output
                yield output

    def _stream_openai(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ) -> Iterable[str]:
        """
        Generate text from a prompt using the model in streaming mode.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        # Check how many messages and contexts we can fit in the model context length constraints
        _, (n_contexts, n_messages_to_keep) = self.format_prompt(
            messages, contexts, **prompt_structure_dict, lang=lang
        )

        messages = self._format_prompt_openai(
            messages=messages[:n_messages_to_keep],  # keep only the last n_messages_to_keep
            contexts=contexts[:n_contexts],  # keep only the last n_contexts
            **prompt_structure_dict,
            lang=lang,
        )

        try:

            response = self.client.chat.completions.create(
                model=self.llm_name,
                messages=messages,
                max_tokens=generation_config_dict["max_new_tokens"],
                temperature=generation_config_dict["temperature"],
                top_p=generation_config_dict["top_p"],
                presence_penalty=generation_config_dict["repetition_penalty"],
                seed=generation_config_dict["seed"],
                n=1,
                stream=True,
            )
            for chunk in response:
                if chunk.choices[0].finish_reason == "stop":
                    return
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[
                        0
                    ].delta.content  # return the delta text message

        except Exception as e:
            yield return_openai_error(e)
            return
            
    def stream(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> str | Iterable[str] | None:
        """
        Generate text from a prompt using the model in streaming mode.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. List of pairs (role, content)
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        stop_words : List[str]
            The stop words to use to stop generation.
        Returns
        str
            The generated text.
        """

        if self.use_openai_api:
            for chunk in self._stream_openai(
                messages=messages,
                contexts=contexts,
                prompt_structure_dict=prompt_structure_dict,
                generation_config_dict=generation_config_dict,
                lang=lang,
                **kwargs,
            ):
                yield chunk
        else:
            for chunk in self._stream_vllm(
                messages=messages,
                contexts=contexts,
                prompt_structure_dict=prompt_structure_dict,
                generation_config_dict=generation_config_dict,
                lang=lang,
                stop_words=stop_words,
                **kwargs,
            ):
                yield chunk


class AsyncVLLMModel(RAGLLM):
    """
    A client that sends asynchronous requests to the VLLM server.
    """

    def __init__(
        self, llm_name: str, base_url: str = None, use_openai_api: bool = True, **kwargs
    ):
        super().__init__(llm_name=llm_name, **kwargs)
        if base_url is None:
            self.endpoint_url = os.environ["VLLM_ENDPOINT_URL"]
        else:
            self.endpoint_url = base_url

        self.llm_name = llm_name

        if use_openai_api:
            self.client = AsyncOpenAI(base_url=self.endpoint_url)  # for VLLM OpenAI compatible API
        self.use_openai_api = use_openai_api
        print(f"Using vLLM OpenAI compatible API server: {self.use_openai_api}")

    def _format_prompt_openai(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        system_prefix: str,
        lang: str = "en",
        **kwargs,
    ) -> List[Dict[str, str]]:
        """
        Formats the prompt to be used by the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : list
            The context to use.
        lang : str, optional
            The language of the prompt, by default 'en'
        """
        system_prompt = self.format_system_prompt(
            contexts=contexts,
            system_prefix=system_prefix,
            lang=lang,
        )

        final_messages = [{"role": "system", "content": system_prompt}] + messages

        return final_messages


    async def _generate_vllm(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> str:
        """
        Generate text from a prompt using the vLLM API server.
        """
        prompt, _ = self.format_prompt(
            messages, contexts, **prompt_structure_dict, lang=lang
        )

        pload = {
            "prompt": prompt,
            "n": 1,
            "use_beam_search": False,
            "top_k": generation_config_dict["top_k"],
            "top_p": generation_config_dict["top_p"],
            "temperature": generation_config_dict["temperature"],
            "max_tokens": generation_config_dict["max_new_tokens"],
            "frequency_penalty": generation_config_dict["repetition_penalty"],
            "stop": stop_words,
            "stream": False,
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(self.endpoint_url, json=pload) as response:
                if response.status != 200:
                    logger.error(
                        f"Error with the request to the vLLM server: {await response.text()}"
                    )
                    raise RequestException()

                data = await response.json()
                output = data["text"][0][len(prompt) :]

                if not output:  # if there is an error vllm returns an empty string
                    logger.error(f"Error with the request to the vLLM server.")
                    raise RequestException()

                return output

    async def _generate_openai(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ) -> str:
        """
        Generate text from a prompt using the vLLM OpenAI API.
        """
        _, (n_contexts, n_messages_to_keep) = self.format_prompt(
            messages, contexts, **prompt_structure_dict, lang=lang
        )

        messages = self._format_prompt_openai(
            messages=messages[:n_messages_to_keep],
            contexts=contexts[:n_contexts],
            **prompt_structure_dict,
            lang=lang,
        )

        try:
            print(f'LLM name: {self.llm_name}')
            response = await self.client.chat.completions.create(
                model=self.llm_name,
                messages=messages,
                max_tokens=generation_config_dict["max_new_tokens"],
                temperature=generation_config_dict["temperature"],
                top_p=generation_config_dict["top_p"],
                presence_penalty=generation_config_dict["repetition_penalty"],
                seed=generation_config_dict["seed"],
                n=1,
                stream=False,
            )
            return response.choices[0].message.content
        except Exception as e:
            return return_openai_error(e)

    async def generate(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> str | None:
        """
        Generate text from a prompt using the model.
        """
        if self.use_openai_api:
            return await self._generate_openai(
                messages=messages,
                contexts=contexts,
                prompt_structure_dict=prompt_structure_dict,
                generation_config_dict=generation_config_dict,
                lang=lang,
                **kwargs,
            )
        else:
            return await self._generate_vllm(
                messages=messages,
                contexts=contexts,
                prompt_structure_dict=prompt_structure_dict,
                generation_config_dict=generation_config_dict,
                lang=lang,
                stop_words=stop_words,
                **kwargs,
            )

    async def _stream_vllm(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> AsyncIterable[str]:
        """
        Generate text from a prompt using the model.
        """
        prompt, _ = self.format_prompt(
            messages, contexts, **prompt_structure_dict, lang=lang
        )

        pload = {
            "prompt": prompt,
            "n": 1,
            "use_beam_search": False,
            "top_k": generation_config_dict["top_k"],
            "top_p": generation_config_dict["top_p"],
            "temperature": generation_config_dict["temperature"],
            "max_tokens": generation_config_dict["max_new_tokens"],
            "frequency_penalty": generation_config_dict["repetition_penalty"],
            "stop": stop_words,
            "stream": True,
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(self.endpoint_url, json=pload, stream=True) as response:
                prev_output = pload["prompt"]
                async for n_token, chunk in enumerate(response.content.iter_chunks()):
                    if n_token == 1 and not output:  # if there is an error vllm returns an empty string as the second chunk (the first one is the prompt)
                        raise RequestException()
                    if chunk:
                        data = json.loads(chunk.decode("utf-8"))
                        if 'detail' in data:
                            logger.error(
                                f"Error with the request to the vLLM server: {data['detail']}"
                            )
                            raise RequestException()
                        output = data["text"]
                        output = output[0][len(prev_output) :]
                        prev_output += output
                        yield output

    async def _stream_openai(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ) -> AsyncIterable[str]:
        """
        Generate text from a prompt using the model in streaming mode.
        """
        _, (n_contexts, n_messages_to_keep) = self.format_prompt(
            messages, contexts, **prompt_structure_dict, lang=lang
        )

        messages = self._format_prompt_openai(
            messages=messages[:n_messages_to_keep],
            contexts=contexts[:n_contexts],
            **prompt_structure_dict,
            lang=lang,
        )

        try:
            response = await self.client.chat.completions.create(
                model=self.llm_name,
                messages=messages,
                max_tokens=generation_config_dict["max_new_tokens"],
                temperature=generation_config_dict["temperature"],
                top_p=generation_config_dict["top_p"],
                presence_penalty=generation_config_dict["repetition_penalty"],
                seed=generation_config_dict["seed"],
                n=1,
                stream=True,
            )
            async for chunk in response:
                if chunk.choices[0].finish_reason == "stop":
                    return
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content

        except Exception as e:
            yield return_openai_error(e)
            return

    async def stream(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> AsyncIterable[str] | None:
        """
        Generate text from a prompt using the model in streaming mode.
        """
        if self.use_openai_api:
            async for chunk in self._stream_openai(
                messages=messages,
                contexts=contexts,
                prompt_structure_dict=prompt_structure_dict,
                generation_config_dict=generation_config_dict,
                lang=lang,
                **kwargs,
            ):
                yield chunk
        else:
            async for chunk in self._stream_vllm(
                messages=messages,
                contexts=contexts,
                prompt_structure_dict=prompt_structure_dict,
                generation_config_dict=generation_config_dict,
                lang=lang,
                stop_words=stop_words,
                **kwargs,
            ):
                yield chunk



    

def return_openai_error(e):
    logger.error(f"Error with the request to the vLLM OpenAI server: {e}")
    if e.code == 400: # BadRequestError
        raise PromptTooLongException()
    elif e.code == 404: # NotFoundError
        raise ModelNotFoundException()
    else:
        logger.error(f"Error with the request to the vLLM OpenAI server: {e.body['message']}")
        raise RequestException()


import aiohttp

class RetrieverClient:
    """Client to retrieve documents from a retriever deployment."""
    def __init__(self, deployment_url):
        self.deployment_url = deployment_url

    async def retrieve(self, query, top_k=5):
        async with aiohttp.ClientSession() as session:
            data = {"query": query, "top_k": top_k}
            async with session.post(self.deployment_url, json=data) as response:
                result = await response.json()
                return result


import os
from logging import getLogger
from typing import List, Dict

from huggingface_hub import hf_hub_download
from ctransformers import AutoModelForCausalLM, AutoConfig

from chat_rag.llms import RAGLLM


logger = getLogger(__name__)


def download_ggml_file(llm_name: str, ggml_model_filename: str, local_path: str):
    """
    Downloads the ggml model file from the huggingface hub.
    Parameters
    ----------
    llm_name : str
        The huggingface repo id.
    ggml_model_filename: str
        The filename of the model to load
    local_path: str
        The local path to save the model file to.
    Returns
    -------
    str
        The path to the downloaded model file.
    """
    logger.info(f"Downloading {ggml_model_filename} from {llm_name}...")
    return hf_hub_download(
        llm_name=llm_name,
        filename=ggml_model_filename,
        local_dir=local_path,
        local_dir_use_symlinks=True,
        token=os.environ["HUGGINGFACE_KEY"],
    )


class GGMLModel(RAGLLM):
    def __init__(self, llm_name: str, ggml_model_filename: str, model_config: str, **kwargs):
        """
        Initializes the ggml model. Optimized for CPU inference
        Parameters
        ----------
        llm_name : str
            The huggingface repo id.
        ggml_model_filename: str
            The filename of the model to load
        model_config: str
            The id of the model config to load
        """
        super().__init__(llm_name, **kwargs)
        local_path = os.path.abspath("models/")
        filename_path = os.path.join(local_path, ggml_model_filename)

        if not os.path.exists(filename_path):
            download_ggml_file(llm_name, ggml_model_filename, local_path)

        logger.info(f"Loading GGML {ggml_model_filename} from {filename_path}...")
        config = AutoConfig.from_pretrained(model_config)
        logger.info("Loaded model config")
        logger.info(config)
        self.model = AutoModelForCausalLM.from_pretrained(
            filename_path,
            config=config,
        )
        logger.info(f"Loaded GGML {ggml_model_filename} from {filename_path}!")

    def generate(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
    ) -> str:
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        stop_words : List[str]
            The stop words to use to stop generation.
        Returns
        -------
        str
            The generated text.
        """

        # if threads not in generation_kwargs, then set threads to half of the available cores
        if (
            "threads" not in generation_config_dict
            or generation_config_dict["threads"] is None
        ):
            generation_config_dict["threads"] = os.cpu_count() // 2

        generation_config_dict["stop"] = stop_words

        prompt = self.format_prompt(messages, contexts, **prompt_structure_dict, lang=lang)

        text = self.model(prompt, **generation_config_dict)
        return text

    def stream(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
    ) -> str:
        """
        Generate text from a prompt using the model in streaming mode.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        stop_words : List[str]
            The stop words to use to stop generation.
        Returns
        -------
        str
            The generated text.
        """

        # if threads not in generation_kwargs, then set threads to half of the available cores
        if(
            "threads" not in generation_config_dict
            or generation_config_dict["threads"] is None
        ):
            generation_config_dict["threads"] = os.cpu_count() // 2

        generation_config_dict["stop"] = stop_words

        prompt = self.format_prompt(messages, contexts, **prompt_structure_dict, lang=lang)

        streamer = self.model(prompt, **generation_config_dict)  # returns a generator
        for word in streamer:
            yield word



from typing import Any, Dict, List

import torch
from sentence_transformers import CrossEncoder

models_dict = {
    'en': 'cross-encoder/ms-marco-MiniLM-L-6-v2', # 'cross-encoder/ms-marco-MiniLM-L-12-v2' for little bit better performance
    'multilingual': 'nreimers/mmarco-mMiniLMv2-L6-H384-v1', # 'nreimers/mmarco-mMiniLMv2-L12-H384-v1' for better performance but slower
}


class ReRanker:
    def __init__(self, lang: str = 'en', device: str = 'cuda', model_name: str = None) -> None:
        """
        Class to rerank the retrieved contexts using a cross-encoder.
        It also filters out low confidence contexts.
        """
        lang = lang if lang in models_dict else 'multilingual' # default to multilingual model if language not supported
        model_name = model_name if model_name is not None else models_dict[lang]
        self.model = CrossEncoder(model_name, max_length=512, device=device)
        self.confidence_threshold = 0.5
        self.activation_fct = torch.sigmoid
        self.max_query_length = 64 # These cross-encoder are not trained on long queries, so we restrict them down to simple queries of max 64 characters 
        # You can see some of these stats here (https://huggingface.co/datasets/ms_marco)

    def __call__(self, query: str, contexts: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Rerank the retrieved contexts using a cross-encoder.
        Parameters
        ----------
        query: str
            The user message.
        contexts: List[Dict[str, Any]]
            List of retrieved contexts.
        Returns
        -------
        List[Dict[str, Any]]
            List of reranked contexts.
        """
        if len(query) > self.max_query_length: # for queries longer than 64 characters, we do not rerank
            return contexts
        
        pairs = [(query, context['content']) for context in contexts]
        scores = self.model.predict(pairs, activation_fct=self.activation_fct)
        print(f"CrossEncoder: {scores} scores")

        # filter out low confidence scores and sort by score
        scores = [(score, context) for score, context in zip(scores, contexts) if score > self.confidence_threshold]
        print(f"CrossEncoder: {len(scores)} contexts left after filtering")

        if len(scores) == 0:
            return []
        
        scores.sort(key=lambda x: x[0], reverse=True)
        return [context for _, context in scores]




from typing import List, Tuple, Dict, Optional
from logging import getLogger

import numpy as np
import torch

from chat_rag.inf_retrieval.embedding_models.base_model import BaseModel

logger = getLogger(__name__)


class SemanticRetriever:
    """
    Class for retrieving the context for a given query using embeddings.
    """

    def __init__(
        self,
        data: Dict[str, List[str]] = None,
        embeddings: Optional[np.ndarray] = None,
        embedding_model: Optional[BaseModel] = None,
    ):
        """
        Parameters
        ----------
        data: Dict[str, List[str]], optional
            Dictionary containing the context to be used for retrieval, by default None
        embeddings: List[np.ndarray], optional
            List of embeddings to be used for retrieval, by default None
        embedding_model: BaseModel, optional
            Embedding model to be used for retrieval, by default None
        """

        # assert that the length of every column is the same
        if data is not None:
            assert (
                len(set([len(data[col]) for col in data])) == 1
            ), "All columns must have the same length"

        self.data = data
        self.len_data = len(data[list(data.keys())[0]]) if data is not None else None
        self.keys_list = list(data.keys()) if data is not None else None

        self.embeddings = (
            torch.from_numpy(embeddings) if embeddings is not None else None
        )

        if embeddings is None:
            logger.info(f"Embeddings not provided.")
        else:
            logger.info(f"Embeddings provided with shape {embeddings.shape}")

        self.embedding_model = embedding_model

    def get_top_matches(
        self,
        queries: List[str],
        top_k: int = 5,
        prefix: str = "query: ",
        threshold: float = None,
        disable_progress_bar: bool = False,
    ) -> List[Tuple[List[float], List[int]]]:
        """
        Returns the top_k most relevant context for the queries.

        Parameters
        ----------
        queries : List[str]
            Queries to be used for retrieval.
        top_k : int, optional
            Number of context to be returned, by default 5. If -1, all context are returned.
        disable_progress_bar : bool, optional
            Whether to disable the progress bar, by default True.
        prefix : str, optional
            Prefix or instruction to be added to the context, by default 'query: ' for e5 models.
        threshold : float, optional
            Minimum score to be returned, by default None.

        Returns
        -------
        list
            List containing tuples of scores and indexes of the context.
        """

        assert hasattr(
            self, "embeddings"
        ), "Embeddings not built. Call build_embeddings() first."

        queries = [prefix + query for query in queries]  # Add query prefix

        queries_embeddings = self.embedding_model.encode(
            queries, disable_progress_bar=disable_progress_bar
        )
        scores = (
            torch.mm(queries_embeddings, self.embeddings.transpose(0, 1)).cpu().numpy()
        )

        sorted_indices = np.argsort(scores, axis=1)[:, ::-1]

        # Use numpy's advanced indexing to get the sorted scores and indices
        sorted_scores = np.take_along_axis(scores, sorted_indices, axis=1)[
            :, :, None
        ]  # Add a new axis to be able to broadcast
        sorted_indexes = np.take_along_axis(
            np.broadcast_to(np.arange(scores.shape[1]), scores.shape),
            sorted_indices,
            axis=1,
        )[
            :, :, None
        ]  # Add a new axis to be able to broadcast

        # If threshold is given, create a mask for scores
        mask = (
            sorted_scores >= threshold
            if threshold is not None
            else np.ones_like(sorted_scores, dtype=bool)
        )

        # Use the mask to filter values
        sorted_scores = np.where(mask, sorted_scores, -np.inf).squeeze(-1)
        sorted_indexes = np.where(mask, sorted_indexes, -1).squeeze(-1)

        # remove the masked values
        sorted_scores = sorted_scores[sorted_scores != -np.inf].reshape(
            len(queries), -1
        )
        sorted_indexes = sorted_indexes[sorted_indexes != -1].reshape(len(queries), -1)

        # If top_k is not -1, truncate results
        if top_k != -1:
            sorted_scores = sorted_scores[:, :top_k]
            sorted_indexes = sorted_indexes[:, :top_k]

        scores_indexes = [
            (score_list, index_list)
            for score_list, index_list in zip(sorted_scores, sorted_indexes)
        ]
        return scores_indexes

    def _get_contexts(
        self, matches: Tuple[List[float], List[int]]
    ) -> List[Dict[str, str]]:
        """
        Returns the context for the matches.
        Parameters
        ----------
        matches : Tuple[List[float], List[int]]
            Tuple containing the scores and the index of the context.
        Returns
        -------
        List[Dict[str, str]]
            List of dictionaries containing the context.
        """
        # Resulting list of dictionaries
        context_list = []

        # Iterate through the matches
        for match in zip(*matches):
            context_dict = {}
            context_dict['similarity'] = match[0]
            # Extract data based on the keys and the matched index
            for key in self.keys_list:  # Using `self.keys_list` here
                context_dict[key] = self.data[key][match[1]]

            context_list.append(context_dict)

        return context_list

    def get_contexts(
        self, matches_batch: List[Tuple[List[float], List[int]]]
    ) -> List[List[Dict[str, str]]]:
        """
        Returns the context for the matches in a batch.
        Parameters
        ----------
        matches_batch : List[List[Tuple[float, int]]]
            List of lists of tuples containing the score and the index of the context.
        Returns
        -------
        List[List[Dict[str, str]]]
            List of lists of dictionaries containing the context.
        """
        return [self._get_contexts(matches) for matches in matches_batch]

    def retrieve(
        self,
        queries: List[str],
        top_k: int = 5,
        prefix: str = "query: ",
        threshold: float = None,
        disable_progress_bar: bool = False,
    ) -> List[List[Dict[str, str]]]:
        """
        Returns the context for the queries.
        Parameters
        ----------
        queries : List[str]
            List of queries to be used for retrieval.
        top_k : int, optional
            Number of context to be returned, by default 5. If -1, all context are returned.
        disable_progress_bar : bool, optional
            Whether to disable the progress bar, by default True.
        prefix : str, optional
            Prefix or instruction to be added to the context, by default 'query: ' for e5 models.
        Returns
        -------
        List[List[Dict[str, str]]]
            List of lists of dictionaries containing the context.
        """

        matches_batch = self.get_top_matches(
            queries,
            top_k=top_k,
            disable_progress_bar=disable_progress_bar,
            prefix=prefix,
            threshold=threshold,
        )

        return self.get_contexts(matches_batch)



import os
from typing import Dict, List

from anthropic import Anthropic, AsyncAnthropic

from chat_rag.llms import CONTEXT_PREFIX, RAGLLM


class ClaudeChatModel(RAGLLM):
    def __init__(self, llm_name, **kwargs) -> None:
        self.llm_name = llm_name
        self.client = Anthropic(
            api_key=os.environ.get("ANTHROPIC_API_KEY"),
        )

    def format_prompt(
        self,
        contexts: List[str],
        system_prefix: str,
        n_contexts_to_use: int = 3,
        lang: str = "en",
        **kwargs,
    ) -> List[Dict[str, str]]:
        """
        Formats the prompt to be used by the model.
        Parameters
        ----------
        contexts : list
            The context to use.
        system_prefix : str
            The prefix to indicate instructions for the LLM.
        system_tag : str
            The tag to indicate the start of the system prefix for the LLM.
        system_end : str
            The tag to indicate the end of the system prefix for the LLM.
        user_tag : str
            The tag to indicate the start of the user input.
        user_end : str
            The tag to indicate the end of the user input.
        assistant_tag : str
            The tag to indicate the start of the assistant output.
        assistant_end : str
            The tag to indicate the end of the assistant output.
        Returns
        -------
        list
            The formatted prompt.
        """
        system_prompt = self.format_system_prompt(
            contexts=contexts,
            system_prefix=system_prefix,
            n_contexts_to_use=n_contexts_to_use,
            lang=lang,
        )

        return system_prompt

    def generate(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ) -> str:
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        system_prompt = self.format_prompt(contexts, **prompt_structure_dict, lang=lang)

        message = self.client.messages.create(
            model=self.llm_name,
            system=system_prompt,
            messages=messages,
            max_tokens=generation_config_dict['max_new_tokens'],
            temperature=generation_config_dict['temperature'],
            top_p=generation_config_dict['top_p'],
            top_k=generation_config_dict['top_k'],
        )

        return message.content[0].text
    
    def stream(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ):
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        system_prompt = self.format_prompt(contexts, **prompt_structure_dict, lang=lang)

        stream = self.client.messages.create(
            model=self.llm_name,
            system=system_prompt,
            messages=messages,
            max_tokens=generation_config_dict['max_new_tokens'],
            temperature=generation_config_dict['temperature'],
            top_p=generation_config_dict['top_p'],
            top_k=generation_config_dict['top_k'],
            stream=True,
        )

        for event in stream:
            if event.type == "content_block_delta":
                yield event.delta.text

        
class AsyncClaudeChatModel(RAGLLM):
    def __init__(self, llm_name, **kwargs) -> None:
        self.llm_name = llm_name
        self.client = AsyncAnthropic(
            api_key=os.environ.get("ANTHROPIC_API_KEY")
        )

    def format_prompt(
        self,
        contexts: List[str],
        system_prefix: str,
        n_contexts_to_use: int = 3,
        lang: str = "en",
        **kwargs,
    ) -> List[Dict[str, str]]:
        """
        Formats the prompt to be used by the model.
        Parameters
        ----------
        contexts : list
            The context to use.
        system_prefix : str
            The prefix to indicate instructions for the LLM.
        system_tag : str
            The tag to indicate the start of the system prefix for the LLM.
        system_end : str
            The tag to indicate the end of the system prefix for the LLM.
        user_tag : str
            The tag to indicate the start of the user input.
        user_end : str
            The tag to indicate the end of the user input.
        assistant_tag : str
            The tag to indicate the start of the assistant output.
        assistant_end : str
            The tag to indicate the end of the assistant output.
        Returns
        -------
        list
            The formatted prompt.
        """
        system_prompt = self.format_system_prompt(
            contexts=contexts,
            system_prefix=system_prefix,
            n_contexts_to_use=n_contexts_to_use,
            lang=lang,
        )

        return system_prompt
    
    async def generate(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ) -> str:
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        system_prompt = self.format_prompt(contexts, **prompt_structure_dict, lang=lang)

        message = await self.client.messages.create(
            model=self.llm_name,
            system=system_prompt,
            messages=messages,
            max_tokens=generation_config_dict['max_new_tokens'],
            temperature=generation_config_dict['temperature'],
            top_p=generation_config_dict['top_p'],
            top_k=generation_config_dict['top_k'],
        )

        return message.content[0].text
    
    async def stream(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        **kwargs,
    ):
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. Pair of (role, message).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        Returns
        -------
        str
            The generated text.
        """

        system_prompt = self.format_prompt(contexts, **prompt_structure_dict, lang=lang)

        stream = await self.client.messages.create(
            model=self.llm_name,
            system=system_prompt,
            messages=messages,
            max_tokens=generation_config_dict['max_new_tokens'],
            temperature=generation_config_dict['temperature'],
            top_p=generation_config_dict['top_p'],
            top_k=generation_config_dict['top_k'],
            stream=True,
        )

        async for event in stream:
            if event.type == "content_block_delta":
                yield event.delta.text



from typing import List, Dict, Optional
import os

from transformers import AutoTokenizer, AutoConfig

from chat_rag.exceptions import PromptTooLongException


CONTEXT_PREFIX = {
    "en": "Information:",
    "fr": "Informations:",
    "es": "Información:",
}

NO_CONTEXT_SUFFIX = {
    "en": "No information provided.",
    "fr": "Aucune information n'a été fournie.",
    "es": "No se proporciona información.",
}




class RAGLLM:
    def __init__(
        self,
        llm_name: str,
        model_max_length: int = None,
        trust_remote_code_tokenizer: bool = False,
        trust_remote_code_model: bool = False,
        **kwargs,
    ) -> None:

        auth_token = os.environ["HUGGINGFACE_KEY"]

        self.tokenizer = AutoTokenizer.from_pretrained(
            llm_name, trust_remote_code=trust_remote_code_tokenizer, token=auth_token
        )

        if model_max_length is not None:
            self.model_max_length = model_max_length
        else:
            self.config = AutoConfig.from_pretrained(
                llm_name, trust_remote_code=trust_remote_code_model, token=auth_token
            )
            self.model_max_length = (
                self.config.max_position_embeddings
                if self.config.max_position_embeddings is not None
                else self.tokenizer.model_max_length
            )

        self.has_chat_template = self.tokenizer.chat_template is not None
        print(f"Model max length: {self.model_max_length}")


    def format_system_prompt(
            self,
            contexts: List[str],
            system_prefix: str,
            lang: str = "en",
            **kwargs,
        ) -> str:
        """
        Formats the system prompt to be used by the model.
        """
        if len(contexts) > 0:
            system_prompt = f"{system_prefix}\n{CONTEXT_PREFIX[lang]}\n"

            for ndx, context in enumerate(contexts):
                system_prompt += f"- {context}"

                if ndx < len(contexts) - 1: # no newline on last context
                    system_prompt += "\n"

            return system_prompt
        else:
            return system_prefix + f"\n{CONTEXT_PREFIX[lang]}\n{NO_CONTEXT_SUFFIX[lang]}"

    def apply_chat_template(self, messages, contexts, system_prefix, system_tag, system_end, user_tag, user_end, assistant_tag, assistant_end, lang):
        """
        Applies the chat template to the messages and contexts.
        """
        system_prompt = self.format_system_prompt(
                contexts, system_prefix, lang
            )

        if self.has_chat_template:
            new_messages = messages.copy() # copy the messages to avoid modifying the original list
            new_messages.insert(0, {"role": "system", "content": system_prompt})
            prompt = self.tokenizer.apply_chat_template(
                    new_messages,
                    add_generation_prompt=True,
                    tokenize=False,
                )
        else:
            if (system_tag == "" or system_tag is None):  # To avoid adding the role_end tag if there is no system prefix
                prompt = f"{system_prompt}\n"
            else:
                prompt = f"{system_tag}{system_prompt}{system_end}"

            for message in messages:
                if message['role'] == 'user':
                    prompt += f"{user_tag}{message['content']}{user_end}{assistant_tag}"
                elif message['role'] == 'assistant':
                    prompt += f"{message['content']}{assistant_end}"
        return prompt

    def format_prompt(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        system_prefix: str,
        system_tag: str,
        system_end: str,
        user_tag: str,
        user_end: str,
        assistant_tag: str,
        assistant_end: str,
        n_contexts_to_use: int = 3,
        lang: str = "en",
        **kwargs,
    ) -> str:
        """
        Formats the prompt to be used by the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. List of pairs (role, content).
        contexts : list
            The context to use.
        system_prefix : str
            The prefix to indicate instructions for the LLM.
        system_tag : str
            The tag to indicate the start of the system prefix for the LLM.
        system_end : str
            The tag to indicate the end of the system prefix for the LLM.
        user_tag : str
            The tag to indicate the start of the user input.
        user_end : str
            The tag to indicate the end of the user input.
        assistant_tag : str
            The tag to indicate the start of the assistant output.
        assistant_end : str
            The tag to indicate the end of the assistant output.
            The tag to indicate the end of the role (system role, user role, assistant role).
        n_contexts_to_use : int, optional
            The number of contexts to use, by default 3
        lang : str, optional
            The language of the prompt, by default 'en'
        """

        n_messages_to_keep = len(messages)
        n_contexts = n_contexts_to_use if len(contexts) > n_contexts_to_use else len(contexts)

        prompt = self.apply_chat_template(messages, contexts, system_prefix, system_tag, system_end, user_tag, user_end, assistant_tag, assistant_end, lang)
        num_tokens = len(self.tokenizer.tokenize(prompt))

        margin = int(self.model_max_length * 0.1)

        while num_tokens > (self.model_max_length -  margin):

            if n_contexts == 1 and n_messages_to_keep > 1:
                n_messages_to_keep -= 1

            n_contexts = n_contexts - 1 if n_contexts > 1 else 1

            # When we reach the minimum number of contexts and messages and the prompt is still too long, we return None
            if n_contexts == 1 and n_messages_to_keep == 1:
                raise PromptTooLongException()

            prompt = self.apply_chat_template(messages[:n_messages_to_keep], contexts[:n_contexts], system_prefix, system_tag, system_end, user_tag, user_end, assistant_tag, assistant_end, lang)

            # get number of tokens
            num_tokens = len(self.tokenizer.tokenize(prompt))

        return prompt, (n_contexts, n_messages_to_keep)

    def generate(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: Dict,
        generation_config_dict: Dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> Optional[str | List[str]]:
        pass

    def stream(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: Dict,
        generation_config_dict: Dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> Optional[str | List[str]]:
        pass



from typing import List

from sklearn.cluster import HDBSCAN # TODO: maybe replace with cuML implementation for speedup when using GPU

from chat_rag.inf_retrieval.embedding_models.base_model import BaseModel

def clusterize_text(queries: List[str], embedding_model: BaseModel, batch_size: int = 32, prefix: str = 'query: '):
    """
    Returns the clusters for the queries.
    Parameters
    ----------
    queries : List[str]
        List of queries to be used for clustering.
    embedding_model: BaseModel
        Embedding model to be used for clustering
    batch_size : int, optional
        Batch size for the embedding model, by default 32.
    prefix : str, optional
        Prefix for the queries, by default 'query: '.
    """
    assert prefix in ['query: ', 'passage: '], "prefix must be 'query: ' or 'passage: '"

    print(f"Generating embeddings for {len(queries)} queries...")
    queries_embeddings = embedding_model.build_embeddings(queries, batch_size=batch_size, prefix=prefix, disable_progress_bar=False) # specific prefix for e5 models queries

    print("Clustering...")
    MIN_CLUSTERS = 3
    max_cluster_size = len(queries) // MIN_CLUSTERS  # at least 3 clusters
    max_cluster_size = max_cluster_size if len(queries) >= (2*MIN_CLUSTERS) else None # if there are less than 6 queries, don't limit the cluster size
    clusterer = HDBSCAN(min_cluster_size=2, max_cluster_size=max_cluster_size, min_samples=1, metric='euclidean')
    clusterer.fit(queries_embeddings)

    return clusterer.labels_



from logging import getLogger
from typing import List, Dict

from chat_rag.llms import RAGLLM

logger = getLogger(__name__)


class RAG:
    """
    Class for generating responses using the Retrieval-Augmented Generation (RAG) pattern.
    """
    def __init__(
        self,
        retriever,
        llm_model: RAGLLM,
        lang: str = "en",
    ):
        """
        Parameters
        ----------
        retriever :
            Retriever object for retrieving contexts.
        llm_model : RAGLLM
            Language model for generating responses.
        lang : str, optional
            Language of the language model, by default "en"
        """

        self.retriever = retriever
        self.model = llm_model
        self.lang = lang


    def retrieve(
        self,
        message: str,
        prev_contents: List[str],
        prompt_structure_dict: dict,
    ):
        """
        Retrieve new contexts if needed.
        Parameters
        ----------
        message : str
            User message.
        prev_contents : List[str]
            List of previous contexts.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        Returns
        -------
        Tuple[List[str], List[Dict[str, str]]]
            List of all conversation contexts and list of the retrieved contexts for the current user message.
        """
        logger.info("Retrieving new contexts")
        contexts = self.retriever.retrieve([message], top_k=prompt_structure_dict["n_contexts_to_use"])[0] # retrieve contexts
        if len(contexts) == 0:
            return [], []

        contents = [context["content"] for context in contexts] # get unique contexts
        returned_contexts = [contexts[:prompt_structure_dict["n_contexts_to_use"]]] # structure for references

        # Use a list comprehension to preserve order and not adding duplicates
        seen = set()
        contents = [x for x in prev_contents + contents if not (x in seen or seen.add(x))]

        return contents, returned_contexts

    def stream(
        self,
        messages: List[Dict[str, str]],
        prev_contents: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict,
        stop_words: List[str] = None,
    ):

        # Retrieve
        contents, returned_contexts = self.retrieve(messages[-1]['content'], prev_contents, prompt_structure_dict)

        # Generate
        for new_text in self.model.stream(
            messages,
            contents,
            prompt_structure_dict=prompt_structure_dict,
            generation_config_dict=generation_config_dict,
            lang=self.lang,
            stop_words=stop_words,
        ):
            yield {
                "res": new_text,
                "context": returned_contexts,
            }

    def generate(
        self,
        messages: List[Dict[str, str]],
        prev_contents: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict,
        stop_words: List[str] = None,
    ):

        # Retrieve
        contents, returned_contexts = self.retrieve(messages[-1]['content'], prev_contents, prompt_structure_dict)

        output_text = self.model.generate(
            messages,
            contents,
            prompt_structure_dict=prompt_structure_dict,
            generation_config_dict=generation_config_dict,
            lang=self.lang,
            stop_words=stop_words,
        )

        return {
            "res": output_text,
            "context": returned_contexts,
        }





from typing import List, Callable
import re
import os

import nltk
from transformers import AutoTokenizer

class WordSplitter:
    """
    Splits a text into chunks of num_words words with an overlap of chunk_overlap words.
    """
    def __init__(self, chunk_size: int = 200, chunk_overlap: int = 20):
        """
        Parameters
        ----------
        chunk_size : int
            The number of words per chunk.
        overlap : int
            The number of words that overlap between two chunks.
        """
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        
    def __call__(self, text: str) -> List[str]:
        """
        Splits a text into chunks of chunk_size words with an overlap of chunk_overlap words.
        Parameters
        ----------
        text : str
            The text to split.
        Returns
        -------
        List[str]
            A list of chunks.
        """
        words = text.split()
        if len(words) <= self.chunk_size:
            return [text]  # Do not split if text has fewer words than chunk_size
        chunks = []
        start = 0
        while start < len(words):
            end = start + self.chunk_size
            chunk = " ".join(words[start:end])
            # check if this chunk is a subset of the previous chunk
            if chunks and chunk in chunks[-1]:
                start += self.chunk_size - self.chunk_overlap
                continue
            chunks.append(chunk)
            start = end - self.chunk_overlap  # Move the start index for the next chunk
        return chunks
    

class CharacterSplitter:
    """
    Splits a text into chunks of num_chars characters with an overlap of overlap characters.
    """
    def __init__(self, num_chars: int, overlap: int):
        """
        Parameters
        ----------
        num_chars : int
            The number of characters per chunk.
        overlap : int
            The number of characters that overlap between two chunks.
        """
        self.num_chars = num_chars
        self.overlap = overlap
        
    def __call__(self, text: str) -> List[str]:
        """
        Splits a text into chunks of num_chars characters with an overlap of overlap characters.
        Parameters
        ----------
        text : str
            The text to split.
        Returns
        -------
        List[str]
            A list of chunks.
        """
        if len(text) <= self.num_chars:
            return [text]  # Do not split if text has fewer characters than num_chars
        chunks = []
        start = 0
        while start < len(text):
            end = start + self.num_chars
            # If we're in the middle of a word, find the next space to end the chunk
            while end < len(text) and text[end] not in [' ', '\n', '\t']:
                end += 1
            chunk = text[start:end]
            # Check if this chunk is a subset of the previous chunk
            if chunks and chunk in chunks[-1]:
                # Adjust start to skip characters equal to specified overlap
                for _ in range(self.overlap):
                    while start < len(text) and text[start] != ' ':
                        start += 1
                    start += 1
                continue
            chunks.append(chunk)
            # Move the start index for the next chunk, ensuring the overlap starts from the beginning of a word
            overlap_chars = 0
            temp_start = end
            while (overlap_chars < self.overlap and temp_start > 0) and temp_start < len(text):
                temp_start -= 1
                if temp_start < 0:  # Avoid going out of bounds
                    break
                if text[temp_start] in [' ', '\n', '\t']:
                    break

            start = temp_start if overlap_chars == self.overlap else end
        return chunks
    

class TokenSplitter:
    """
    Splits a text into chunks of n xºtokens with an overlap of overlap tokens.
    Inspired by https://github.com/jerryjliu/llama_index/blob/main/llama_index/text_splitter/token_splitter.py
    """
    def __init__(self, tokenizer_name="intfloat/e5-small-v2", chunk_size: int = 128, chunk_overlap: int = 16, separators: List[str] = ["\n", "\t", " "]):
        """
        Parameters
        ----------
        tokenizer_name : str
            The name of the tokenizer to use.
        chunk_size : int
            The number of tokens per chunk.
        chunk_overlap : int
            The number of tokens that overlap between two chunks.
        separators : List[str]
            A list of separators to split on.
        """
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name) 
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.separators = separators
        self.sep_pattern = '(?={})'.format('|'.join(map(re.escape, separators))) # regex pattern to split on separators

    def __call__(self, text):
        """
        Splits a text into chunks of n xºtokens with an overlap of overlap tokens.
        Parameters
        ----------
        text : str
            The text to split.
        Returns
        -------
        List[str]
            A list of chunks.
        """
        # First, we split using our regular expression pattern
        temp_split = re.split(self.sep_pattern, text)

        # Then, we group consecutive separators with the next word
        split_text = []
        buffer = ""
        for item in temp_split:
            if item in self.separators:
                buffer += item
            else:
                split_text.append(buffer + item)
                buffer = ""
        
        chunks = []
        current_chunk = []
        current_length = 0
        for split in split_text:
            n_tokens = len(self.tokenizer.tokenize(split))

            if current_length + n_tokens > self.chunk_size:
                chunks.append("".join(current_chunk).strip())
                # start a new chunk with overlap
                # keep popping off the first element until we have enough space
                while current_length > self.chunk_overlap or current_length + n_tokens > self.chunk_size:
                    # pop off the first element
                    first_chunk = current_chunk.pop(0)
                    current_length -= len(self.tokenizer.tokenize(first_chunk))
                    
            current_chunk.append(split)
            current_length += n_tokens

        if len(current_chunk) > 0:
            chunks.append("".join(current_chunk).strip())

        return chunks
    

class SentenceTokenSplitter:
    """
    Splits a text into chunks of sentences according to the number of tokens.
    """
    def __init__(self, tokenizer_name="intfloat/e5-small-v2", chunk_size: int = 128):
        """
        Parameters
        ----------
        tokenizer_name : str
            The name of the tokenizer to use.
        chunk_size : int
            The number of tokens per chunk.
        """
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name) 
        self.chunk_size = chunk_size

    def __call__(self, text):
        """
        Splits a text into chunks of sentences according to the number of tokens.
        Parameters
        ----------
        text : str
            The text to split.
        Returns
        -------
        List[str]
            A list of chunks.
        """

        # First, we split the text into sentences
        sentences = nltk.sent_tokenize(text)  
        
        chunks = []
        current_chunk = []
        current_length = 0
        for sentence in sentences:
            n_tokens = len(self.tokenizer.tokenize(sentence))

            if current_length + n_tokens > self.chunk_size and len(current_chunk) > 0: # if the current sentence does not fit in the current chunk, start a new chunk
                chunks.append(" ".join(current_chunk).strip())
                current_chunk = []
                current_length = 0
                    
            current_chunk.append(sentence)
            current_length += n_tokens

        if len(current_chunk) > 0: # add the last chunk
            chunks.append(" ".join(current_chunk).strip())

        return chunks
    

class SmartSplitter:
    """
    Splits the text into information meaningful chunks using the GPT-4 model.
    This can reach API rate limits very quickly.
    """
    def __init__(self, model_name='gpt-4-0125-preview'):
        """
        Parameters
        ----------
        model_name : str
            The name of the model to use. The recommended model is 'gpt-4', do not guarantee that other models will work.
        """
        from openai import OpenAI
        
        self.client = OpenAI(api_key=os.environ["OPENAI_API_KEY"])

        self.model_name = model_name
        
        self.system_prompt = """You are designed to function as an informative assistant. 
- When provided with user input, your primary task is to discern and segregate the text into distinct segments based on their inherent intents. 
- It's essential that each of these segments retains enough context to be useful for future information retrieval tasks. 
- Avoid creating overly short chunks, as they may lack necessary context. 
- If the user input is succinct and doesn't warrant segmentation, there's no obligation to segment it. 
- To ensure easy parsing during subsequent processing, separate each of these segmented chunks with the delimiter '###'."""

    def __call__(self, text):
        """
        Splits the text into information meaningful chunks using the GPT-4 model.
        Parameters
        ----------
        text : str
            The text to split.
        Returns
        -------
        List[str]
            A list of chunks.
        """
        

        messages = [
            {"role": "system", "content": self.system_prompt},
            {"role": "user", "content": text},
        ]

        # Splitting the text into chunks
        response = self.client.chat.completions.create(model=self.model_name,
        messages=messages,
        temperature=0.2,
        max_tokens=1024,
        top_p=0.0,
        presence_penalty=0.0)

        messages.append({"role": "assistant", "content": response.choices[0]["message"]["content"]})
        chunks = response.choices[0]["message"]["content"].split("###")

        # Analyzing each chunk
        for ndx, chunk in enumerate(chunks):
            print(f"Chunk {ndx} of {len(chunks)}", flush=True)
            user_content = f"Analyze this chunk:\n{chunk}" if ndx else f"Reflect upon the answer you produced based on the previous system instruction. Provide feedback on the performance, mentioning any areas of improvement or points of adherence. It's very important that each chunk can be understood on its own. Let's go chunk by chunk.\nAnalyze this chunk:\n{chunk}"
            messages.append({"role": "user", "content": user_content})

            response = self.client.chat.completions.create(model=self.model_name,
            messages=messages,
            temperature=1,
            max_tokens=1024,
            top_p=0.0,
            presence_penalty=0.0)
            messages.append({"role": "assistant", "content": response.choices[0]["message"]["content"]})

        # Refining chunks based on feedback
        messages.append(
            {"role": "user", "content": "Refine your chunks given the feedback that you provided. Remember that a segmentation is not always needed and just write the chunks."}
        )

        response = self.client.chat.completions.create(model=self.model_name,
        messages=messages,
        temperature=0.2,
        max_tokens=1024,
        top_p=0.0,
        presence_penalty=0.0)

        final_chunks = response.choices[0]["message"]["content"].split("###")

        return final_chunks
    

def get_splitter(splitter, chunk_size, chunk_overlap):
    """
    Returns the splitter object corresponding to the splitter name.
    """

    # check if chunk_size and chunk_overlap are valid and that the chunk_overlap is smaller than the chunk_size
    if chunk_size < 1:
        raise ValueError(f"chunk_size must be >= 1, got {chunk_size}")
    if chunk_overlap < 0:
        raise ValueError(f"chunk_overlap must be >= 0, got {chunk_overlap}")
    if chunk_overlap >= chunk_size:
        raise ValueError(f"chunk_overlap must be smaller than chunk_size, got chunk_overlap={chunk_overlap} and chunk_size={chunk_size}")
    
    if splitter == 'sentences':
        return SentenceTokenSplitter(chunk_size=chunk_size)
    elif splitter == 'words':
        return WordSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    elif splitter == 'tokens':
        return TokenSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    elif splitter == 'smart':
        return SmartSplitter()
    else:
        raise ValueError(f"Unknown splitter: {splitter}, must be one of: sentences, words, tokens, smart")


from typing import List
from logging import getLogger

from tqdm import tqdm
import torch
import torch.nn.functional as F
from torch import Tensor
from transformers import AutoTokenizer, AutoModel

logger = getLogger(__name__)


class BaseModel:
    """
    Class for generating embeddings for a given text.
    """

    def __init__(
        self,
        model_name: str = "sentence-transformers/multi-qa-MiniLM-L6-cos-v1",
        use_cpu: bool = False,
        huggingface_key: str = None,
    ):
        """
        Parameters
        ----------
        model_name : str, optional
            Name of the model to be used for encoding the context, by default 'intfloat/multilingual-e5-base'
        use_cpu : bool, optional
            Whether to use CPU for encoding, by default False
        huggingface_key : str, optional
            Huggingface key to be used for private models, by default None
        """

        self.device = "cuda" if (not use_cpu and torch.cuda.is_available()) else "cpu"

        logger.info(f"Using device {self.device}")
        logger.info(f"Loading model {model_name}")

        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name, token=huggingface_key
        )

        self.model = AutoModel.from_pretrained(
            model_name,
            token=huggingface_key,
        ).to(self.device)

    def average_pool(
        self, last_hidden_states: Tensor, attention_mask: Tensor
    ) -> Tensor:
        """
        Average pooling strategy.
        Parameters
        ----------
        last_hidden_states : Tensor
            Last hidden states of the model.
        attention_mask : Tensor
            Attention mask of the model.
        Returns
        -------
        Tensor
            Embeddings of the queries.
        """
        last_hidden = last_hidden_states.masked_fill(
            ~attention_mask[..., None].bool(), 0.0
        )
        return last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]

    def encode(
        self,
        queries: List[str],
        batch_size: int = -1,
        disable_progress_bar: bool = False,
    ) -> torch.Tensor:
        """
        Returns the embeddings of the queries.
        Parameters
        ----------
        queries : List[str]
            List of queries to be encoded.
        batch_size : int, optional
            Batch size, by default -1 (no batching).
        disable_progress_bar : bool, optional
            Whether to disable the progress bar, by default False.
        Returns
        -------
        torch.Tensor
            Embeddings of the queries.
        """
        if batch_size == -1:
            batch_size = len(queries)

        all_embeddings = torch.tensor([])

        # Compute token embeddings
        with torch.inference_mode():
            for i in tqdm(
                range(0, len(queries), batch_size), disable=disable_progress_bar
            ):
                # Tokenize sentences
                encoded_input = self.tokenizer(
                    queries[i : i + batch_size],
                    padding=True,
                    truncation=True,
                    return_tensors="pt",
                ).to(self.device)
                inputs = {key: val for key, val in encoded_input.items()}

                model_output = self.model(**inputs, return_dict=True)

                embeddings = self.average_pool(
                    model_output.last_hidden_state, inputs["attention_mask"]
                )

                del model_output

                embeddings = F.normalize(embeddings, p=2, dim=1).cpu()
                all_embeddings = torch.cat((all_embeddings, embeddings))

        torch.cuda.empty_cache()

        return all_embeddings
    
    def build_embeddings(
        self,
        contents: List[str] = None,
        batch_size: int = 1,
        prefix: str = "passage: ",
        disable_progress_bar: bool = False,
    ):
        """
        Builds the embeddings for the context.
        Parameters
        ----------
        contents : List[str], optional
            List of contents to be encode, by default None
        batch_size : int, optional
            Batch size to be used for encoding the context, by default 1
        prefix : str, optional
            Prefix or instruction to be added to the context, by default 'passage: ' for e5 models.
        """
        logger.info("Building embeddings...")

        contents = [prefix + content for content in contents]  # add prefix to answers
        embeddings = self.encode(contents, batch_size, disable_progress_bar)

        return embeddings



from threading import Thread
from logging import getLogger
from typing import List, Dict
import regex
from collections.abc import Sequence
import os

import torch
from torch import Tensor
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    TextIteratorStreamer,
    StoppingCriteria,
    StoppingCriteriaList,
)

from chat_rag.llms import RAGLLM

logger = getLogger(__name__)


class StopPatternCriteria(StoppingCriteria):
    """
    Stopping criteria that checks if the text generation contains a stop pattern.
    Inspired by https://github.com/microsoft/guidance/blob/e3c6fe93fa00cb86efc130bbce22aa29100936d4/guidance/llms/_transformers.py#L567
    """

    def __init__(self, stop_pattern, tokenizer, prefix_length):
        if isinstance(stop_pattern, str):
            self.stop_patterns = [regex.compile(stop_pattern)]
        else:
            self.stop_patterns = [
                regex.compile(pattern.strip()) for pattern in stop_pattern
            ]
        self.tokenizer = tokenizer
        self.prefix_length = prefix_length

    def __call__(self, input_ids, scores):
        # handle 1D inputs
        if not isinstance(input_ids[0], Sequence) and not (
            hasattr(input_ids[0], "shape") and len(input_ids[0].shape) > 0
        ):
            input_ids = [input_ids]

        text_generations = self.tokenizer.batch_decode(input_ids)

        # check if all text generations have a stop pattern, so we can stop the batch inference
        all_done = True
        for text_generation in text_generations:
            found = False
            for stop_pattern in self.stop_patterns:
                if stop_pattern.search(text_generation[self.prefix_length :].strip()):
                    found = True
            if not found:
                all_done = False
                break

        return all_done


class HFModel(RAGLLM):
    MAX_GPU_MEM = "18GiB"  # Why this
    MAX_CPU_MEM = "12GiB"

    def __init__(
        self,
        llm_name: str,
        use_cpu: bool,
        load_in_8bit: bool = False,
        use_fast_tokenizer: bool = True,
        trust_remote_code_tokenizer: bool = False,
        trust_remote_code_model: bool = False,
        revision: str = "main",
        **kwargs,
    ):
        """
        Initializes the model and tokenizer.
        Parameters
        ----------
        llm_name : str
            The huggingface repo id.
        use_cpu : bool
            Whether to use cpu or gpu.
        load_in_8bit: bool
            Whether to load the model in 8bit mode
        use_fast_tokenizer: bool
            Whether to use the fast tokenizer
        trust_remote_code_tokenizer: bool
            Whether to trust the remote code when loading the tokenizer
        trust_remote_code_model: bool
            Whether to trust the remote code when loading the model
        revision: str
            The specific model version to use. It can be a branch name, a tag name, or a commit id, since we use a git-based system for storing models
        """
        super().__init__(
            llm_name,
            trust_remote_code_tokenizer=trust_remote_code_tokenizer,
            trust_remote_code_model=trust_remote_code_model,
            **kwargs,
        )
        self.use_cpu = use_cpu
        auth_token = os.environ["HUGGINGFACE_KEY"]

        device_map = (
            "auto" if (not use_cpu and torch.cuda.is_available()) else None
        )  # use gpu if available
        self.device = (
            "cuda:0" if (not use_cpu and torch.cuda.is_available()) else None
        )  # For moving tensors to the GPU
        memory_device = {"cpu": self.MAX_CPU_MEM}
        if not use_cpu and torch.cuda.is_available():
            memory_device = {0: self.MAX_GPU_MEM}

        logger.info(f"Loading HF model from {llm_name}...")
        self.tokenizer = AutoTokenizer.from_pretrained(
            llm_name,
            token=auth_token,
            use_fast=use_fast_tokenizer,
            revision=revision,
            trust_remote_code=trust_remote_code_tokenizer,
        )
        self.model = AutoModelForCausalLM.from_pretrained(
            llm_name,
            device_map=device_map,
            torch_dtype="auto",
            max_memory=memory_device,
            low_cpu_mem_usage=True,
            token=auth_token,
            revision=revision,
            load_in_8bit=load_in_8bit,
            trust_remote_code=trust_remote_code_model,
        )
        self.streamer = TextIteratorStreamer(
            self.tokenizer, skip_special_tokens=True, skip_prompt=True
        )

    def generate(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> str:
        """
        Generate text from a prompt using the model.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. List of pairs (role, content).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        stop_words : List[str]
            The stop words to use to stop generation.
        Returns
        -------
        str
            The generated text.
        """
        torch.manual_seed(generation_config_dict["seed"])

        generation_config_dict.pop("seed")

        prompt = self.format_prompt(messages, contexts, **prompt_structure_dict, lang=lang)

        input_ids = self.tokenizer(prompt, return_tensors="pt").input_ids.to(
            self.device
        )

        if stop_words is not None:
            stopping_criteria = StopPatternCriteria(
                stop_words, self.tokenizer, len(prompt)
            )
            generation_config_dict["stopping_criteria"] = StoppingCriteriaList(
                [stopping_criteria]
            )

        generation_config_dict = dict(
            input_ids=input_ids,
            do_sample=True,
            **generation_config_dict,
            eos_token_id=self.tokenizer.eos_token_id,
            pad_token_id=self.tokenizer.pad_token_id,
        )

        generation_config_dict.pop("name")

        with torch.inference_mode():
            outputs = self.model.generate(**generation_config_dict)
            outputs = outputs[:, len(input_ids[0]) :]  # Remove the prompt
        if self.device is not None:
            torch.cuda.empty_cache()

        output = self.tokenizer.decode(outputs[0], skip_special_tokens=True)
        # remove all stop words from output TODO: This is a hack, we should improve the stopping criteria, but it's not easy when one stop word is longer than one token
        if stop_words is not None:
            for stop_word in stop_words:
                output = output.replace(stop_word.strip(), "")

        return output

    def stream(
        self,
        messages: List[Dict[str, str]],
        contexts: List[str],
        prompt_structure_dict: dict,
        generation_config_dict: dict = None,
        lang: str = "en",
        stop_words: List[str] = None,
        **kwargs,
    ) -> str:
        """
        Generate text from a prompt using the model in streaming mode.
        Parameters
        ----------
        messages : List[Tuple[str, str]]
            The messages to use for the prompt. List of pairs (role, content).
        contexts : List[str]
            The contexts to use for generation.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.
        generation_config_dict : dict
            Keyword arguments for the generation.
        lang : str
            The language of the prompt.
        stop_words : List[str]
            The stop words to use to stop generation.
        Returns
        -------
        str
            The generated text.
        """

        torch.manual_seed(generation_config_dict["seed"])

        generation_config_dict.pop("seed")

        prompt = self.format_prompt(messages, contexts, **prompt_structure_dict, lang=lang)

        input_ids = self.tokenizer(prompt, return_tensors="pt").input_ids.to(
            self.device
        )

        generation_config_dict = dict(
            input_ids=input_ids,
            do_sample=True,
            **generation_config_dict,
            eos_token_id=self.tokenizer.eos_token_id,
            pad_token_id=self.tokenizer.pad_token_id,
        )

        generation_config_dict.pop("name")

        generation_config_dict["streamer"] = self.streamer
        thread = Thread(target=self.model.generate, kwargs=generation_config_dict)
        thread.start()
        for new_text in self.streamer:
            if (
                new_text in stop_words
            ):  # TODO: This is a hack, we should use the stopping criteria, but it's not easy when one stop word is longer than one token
                thread.join()
                break
            yield new_text



from typing import List

from openai import OpenAI


def generate_intents(
    intent_queries_list: List[List[str]], max_new_tokens=100
) -> List[str]:
    """
    Generate a batch of texts from a given instruction and a list of text inputs.
    Parameters
    ----------
    intent_queries_list : List[List[str]]
        List containing lists of intent queries.
    max_new_tokens : int, optional
        Maximum number of tokens to be generated, by default 100
    Returns
    -------
    List[str]
        List of generated intents.
    """
    client = OpenAI()

    instruction = "Summarize the intent that represents all the following questions in a few words:",

    # Generate prompts for each intent
    input_texts = []
    for intent_queries in intent_queries_list:
        intent_queries = list(set(intent_queries))
        input_text = f"{instruction}\n"

        # Append each query
        for query in intent_queries[:10]:
            input_text += f"- {query}\n"

        input_text += "The intent is to "
        input_texts.append(input_text)

    responses = []
    #batches of 20
    for i in range(0, len(input_texts), 20):
        result = client.completions.create(model='gpt-3.5-turbo-instruct',
        prompt=input_texts[i:i+20],
        temperature=0.5,
        max_tokens=max_new_tokens).choices
        responses.extend(result)

    return ['To ' + response.text for response in responses]


"""
Given a list of questions and a list of contents, this script computes the similarity between each question and each content ussing a cross encoder.
The cross encoder is a transformer model that takes a pair of sentences and outputs a score between 0 and 1 indicating how similar the sentences are.
If the score is greater than 0.5, the question is considered answerable.
The goal of this script is to get the number of answerable questions and the percentage of answerable questions given a list of contents.
"""
import argparse
import pandas as pd
import numpy as np
import torch
from sentence_transformers import CrossEncoder
from tqdm import tqdm
import os

def get_answerable_questions(contents_path, questions_path, language, device, batch_size, max_questions=None):
    if language == 'en':
        model_name = 'cross-encoder/ms-marco-MiniLM-L-6-v2'
    else:
        model_name = 'nreimers/mmarco-mMiniLMv2-L6-H384-v1'

    model = CrossEncoder(model_name, max_length=512, device=device)
    model.model = torch.compile(model.model, mode='default')

    df = pd.read_csv(contents_path)
    df_q = pd.read_csv(questions_path)

    # sample questions if max_questions is specified
    if max_questions is not None:
        df_q = df_q.sample(max_questions, random_state=42)

    contents = df['content'].tolist()
    questions = df_q['question'].tolist()

    progress_file = questions_path.replace('.csv', '_answerable.npy')
    start_index = 0

    # Check if previous progress exists
    if os.path.exists(progress_file):
        progress_data = np.load(progress_file, allow_pickle=True)
        start_index, scores = progress_data[0], progress_data[1]
    else:
        scores = np.zeros((len(questions), len(contents)))

    len_q = len(questions)
    for i in tqdm(range(start_index, len_q, batch_size)):
        batch_questions = questions[i:i+batch_size]
        pairs = [[q, c] for q in batch_questions for c in contents]
        result = model.predict(pairs, activation_fct=torch.sigmoid)
        result = result.reshape(len(batch_questions), len(contents))
        scores[i:i+batch_size] = result
        np.save(progress_file, np.array([i + batch_size, scores], dtype=object))  # Save current index and scores
        torch.cuda.empty_cache()

    # Post-processing
    df_q['score'] = scores.max(axis=1)
    df_q['answerable'] = df_q['score'] > 0.5
    df_q_path = questions_path.replace('.csv', '_answerable.parquet')
    df_q.to_parquet(df_q_path)

    # remove progress file
    os.remove(progress_file)

    answerable_questions = df_q['answerable'].sum()
    return answerable_questions, answerable_questions / len(df_q)

def main():
    parser = argparse.ArgumentParser(description='Process content and question files to find answerable questions.')
    parser.add_argument('contents_path', type=str, help='Path to the contents CSV file')
    parser.add_argument('questions_path', type=str, help='Path to the questions CSV file')
    parser.add_argument('--lang', type=str, default='en', help='Language of the contents and questions')
    parser.add_argument('--device', type=str, default='cpu', choices=['cuda', 'cpu'], help='Device to run the model on (cuda or cpu)')
    parser.add_argument('--batch_size', type=int, default=4, help='Batch size for processing')
    # add a max_questions argument to limit the number of questions to process, if not specified, process all questions
    parser.add_argument('--max_questions', type=int, default=None, help='Max number of questions to process')

    args = parser.parse_args()

    answerable_count, answerable_percentage = get_answerable_questions(
        args.contents_path, args.questions_path, args.lang, args.device, args.batch_size, args.max_questions)

    print(f"Number of answerable questions: {answerable_count}")
    print(f"Percentage of answerable questions: {answerable_percentage:.2f}")

if __name__ == '__main__':
    main()




from logging import getLogger

import torch
from transformers import AutoTokenizer, AutoModel

from chat_rag.inf_retrieval.embedding_models.base_model import BaseModel

logger = getLogger(__name__)


class E5Model(BaseModel):
    """
    Class for retrieving the context for a given query using embeddings.
    """

    def __init__(
        self,
        model_name: str = "intfloat/e5-small-v2",
        use_cpu: bool = False,
        huggingface_key: str = None,
    ):
        """
        Parameters
        ----------
        model_name : str, optional
            Name of the model to be used for encoding the context, by default 'intfloat/multilingual-e5-base'
        use_cpu : bool, optional
            Whether to use CPU for encoding, by default False
        huggingface_key : str, optional
            Huggingface key to be used for private models, by default None
        """

        self.device = "cuda" if (not use_cpu and torch.cuda.is_available()) else "cpu"

        logger.info(f"Using device {self.device}")
        logger.info(f"Loading model {model_name}")

        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name, token=huggingface_key
        )

        if "e5-small-v2" in model_name:
            self.tokenizer.model_max_length = 512  # Max length not defined on the tokenizer_config.json for the small model

        self.model = AutoModel.from_pretrained(
            model_name,
            token=huggingface_key,
        ).to(self.device)



from pydantic import BaseModel, Field
from datetime import date
from typing import List, Optional, Dict


class KnowledgeItemImage(BaseModel):
    """
    An image contained in a KnowledgeItem.
    """

    image_base64: str
    image_mime_type: str
    image_caption: Optional[str] = None


class KnowledgeItem(BaseModel):
    """
    A KnowledgeItem is a set of elements (titles, texts, urls, etc.) that are used for Retrieval Augmented Generation (RAG).
    """

    content: str
    title: Optional[str] = None
    section: Optional[str] = None
    url: Optional[str] = None
    page_number: Optional[int] = 0
    metadata: Optional[Dict] = None
    images: Optional[Dict[int, KnowledgeItemImage]] = Field(
        default_factory=dict
    )  # Dict of images with their index as key



from typing import List, Optional, Union, BinaryIO, IO, Callable
from tempfile import SpooledTemporaryFile

from unstructured.documents.elements import (
    Element,
    Text,  # Uncategorized text
    Title,
)

from unstructured.partition.auto import partition_pdf, partition_html

from chat_rag.data.models import KnowledgeItem


def is_strict_instance(obj, class_type):
    return isinstance(obj, class_type) and obj.__class__ is class_type


def is_header_or_footer(element: Element) -> bool:
    """
    Checks if an element is a header or footer.
    Parameters
    ----------
    element : Element
        An element to check.
    Returns
    -------
    bool
        True if the element is a header or footer, False otherwise.
    """
    # Check if the type matches typical header/footer types
    element = element.to_dict()
    if element["type"] not in ["Title", "UncategorizedText"]:
        return False

    # Get the Y-coordinate of the top left and bottom left of the potential header/footer
    y_top_left = element["metadata"]["coordinates"]["points"][0][1]
    y_bottom_left = element["metadata"]["coordinates"]["points"][1][1]

    # Page's height for reference
    page_height = element["metadata"]["coordinates"]["layout_height"]

    # Check if the element is within the top 10%, page header
    if y_top_left < 0.1 * page_height or y_bottom_left < 0.1 * page_height:
        return True
    # Check if the element is within the bottom 10% and the text is just a number, page footer
    if (
        y_top_left > 0.9 * page_height or y_bottom_left > 0.9 * page_height
    ) and element["text"].isdigit():
        return True

    return False


def parse_elements(
    elements: List[Element],
    file_type: str = "pdf",
    combine_section_under_n_chars: int = 500,
    new_after_n_chars: int = 1000,
) -> List[List[Element]]:
    """
    Parse a list of elements into sections following a set of rules based on the titles and length parameters.
    Parameters
    ----------
    elements : List[Element]
        A list of unstructured elements. Usually the ouput of a partition functions.
    file_type: str
        The type of file to parse. Can be 'pdf' or 'html'.
    combine_under_n_chars: int
        Combines elements (for example a series of titles) until a section reaches
        a length of n characters.
    new_after_n_chars: int
        Cuts off new sections once they reach a length of n characters
    Returns
    -------
    List[List[Element]]
        A list of sections, where each section is a list of elements.
    """

    sections = []
    section = []
    for element in elements:
        append = False

        if file_type == "pdf" and is_header_or_footer(element):
            continue

        if len(section) == 0:  # start of document
            append = True
        elif isinstance(section[-1], Title) and isinstance(
            element, Title
        ):  # Title -> Title : merge
            append = True
        elif is_strict_instance(section[-1], Text) and isinstance(
            element, Title
        ):  # Uncat Text -> Title : merge
            append = True
        elif isinstance(section[-1], Title) and not isinstance(
            element, Title
        ):  # Title -> Other : merge
            append = True
        elif not isinstance(section[-1], Title) and isinstance(
            element, Title
        ):  # Not Title -> Title : new section
            append = False
        elif not isinstance(section[-1], Title) and not isinstance(
            element, Title
        ):  # Not Title -> Not Title: merge
            append = True
        else:
            append = True

        # if the section is too short, just append
        section_length = sum([len(str(element)) for element in section])
        element_length = len(str(element))
        if (
            combine_section_under_n_chars != -1
            and (section_length + element_length) < combine_section_under_n_chars
        ):
            append = True
        if (
            new_after_n_chars != -1
            and (section_length + element_length) > new_after_n_chars
        ):
            append = False

        if append:
            section.append(element)
        else:
            sections.append(section)
            section = [element]

    sections.append(section)  # append the last section

    return sections


def transform_to_k_items(sections: List[List[Element]], file_type: str = 'pdf',) -> List[KnowledgeItem]:
    """
    Transforms a list of sections into a list of KnowledgeItems.
    Parameters
    ----------
    sections : List[List[KnowledgeItem]]
        A list of sections, where each section is a list of elements.
    file_type: str
        The type of file that was used to generate the sections.
    Returns
    -------
    List[KnowledgeItem]
        A list of KnowledgeItems.
    """

    sections_k_items = []
    prev_title = None
    for ndx, section in enumerate(sections):
        title = None
        for element in section:
            if isinstance(element, Title):
                title = element.text
                break
        if title is None: # if no title is found take the previous title
            title = prev_title
            if ndx == 0: # first section and no title found
                # first 5 words as title
                title = " ".join([word for word in section[0].text.split()[:5]])


        section_k_items = KnowledgeItem(content="\n".join([element.text for element in section]), title=title)
        if file_type == 'pdf':
            section_k_items.page_number = section[0].metadata.page_number
        elif file_type == 'html':
            url = section[0].metadata.url if section[0].metadata.url else section[0].metadata.filename # use url if available, otherwise filename
            section_k_items.url = url

        if section_k_items.content.strip() != "":
            sections_k_items.append(section_k_items)

        prev_title = title # save title for next section

    return sections_k_items


def split_k_items(k_items: List[KnowledgeItem], split_function: Callable = lambda x: [x]) -> List[KnowledgeItem]:
    """
    Splits a list of knowledge items into a list of more knowledge items when the content is split if needed.
    Parameters
    ----------
    k_items : List[KnowledgeItem]
        A list of KnowledgeItems.
    split_function: Callable
        A function that takes a knowledge item and returns a list of knowledge items. The default does not split.
    Returns
    -------
    List[KnowledgeItem]
        A list of KnowledgeItems.
    """

    new_k_items = []
    for k_item in k_items:
        text_splitted = split_function(k_item.content)
        for text in text_splitted:
            c = KnowledgeItem(content=text, title=k_item.title, url=k_item.url, section=k_item.section, page_number=k_item.page_number)
            new_k_items.append(c)

    return new_k_items


def parse_pdf(
    filename: str = "",
    file: Optional[Union[BinaryIO, SpooledTemporaryFile]] = None,
    strategy: str = "auto",
    combine_section_under_n_chars: int = 500,
    new_after_n_chars: int = -1,
    split_function: Callable = lambda x: [x],
) -> List[KnowledgeItem]:
    """
    Parse a pdf file into sections.
    Parameters
    ----------
    filename : str
        The path to the pdf file.
    file : Optional[Union[BinaryIO, SpooledTemporaryFile]]
        The file object of the pdf file.
    strategy : str
        The strategy to use to parse the pdf file. Can be 'auto', 'fast', 'ocr' or 'high_res'.
    combine_section_under_n_chars: int
        Combines elements (for example a series of titles) until a section reaches
        a length of n characters.
    new_after_n_chars: int
        Cuts off new sections once they reach a length of n characters
    split_function: Callable
        A function that takes a knowledge item and returns a list of knowledge items. The default does not split.
    Returns
    -------
    List[KnowledgeItem]
        A list of KnowledgeItem.
    """

    if strategy in ['ocr_only', 'hi_res']:
        print(f'Using strategy {strategy}. This might take a few minutes.')

    elements = partition_pdf(filename=filename, file=file, strategy=strategy)
    print(f"N elements: {len(elements)}")

    sections = parse_elements(
        elements,
        file_type="pdf",
        combine_section_under_n_chars=combine_section_under_n_chars,
        new_after_n_chars=new_after_n_chars,
    )

    print(f"N sections: {len(sections)}")

    k_items = transform_to_k_items(sections, file_type="pdf")

    print(f"N k_items: {len(k_items)}")

    k_items = split_k_items(k_items, split_function=split_function)

    print(f"N k_items after split: {len(k_items)}")

    return k_items


def parse_html(
    filename: Optional[str] = None,
    file: Optional[IO[bytes]] = None,
    text: Optional[str] = None,
    url: Optional[str] = None,
    encoding: Optional[str] = None,
    combine_section_under_n_chars: int = 500,
    new_after_n_chars: int = -1,
    split_function: Callable = lambda x: [x],
) -> List[KnowledgeItem]:
    """
    Parse an html file into sections.
    Parameters
    ----------
    filename : Optional[str]
        The path to the html file.
    file : Optional[IO[bytes]]
        The file object of the html file.
    text : Optional[str]
        The text of the html file.
    url : Optional[str]
        The url of the html file.
    encoding : Optional[str]
        The encoding of the html file.
    combine_section_under_n_chars: int
        Combines elements (for example a series of titles) until a section reaches
        a length of n characters.
    new_after_n_chars: int
        Cuts off new sections once they reach a length of n characters
    split_function: Callable
        A function that takes a knowledge item and returns a list of knowledge items. The default does not split.
    Returns
    -------
    List[KnowledgeItem]
        A list of KnowledgeItems.
    """
    elements = partition_html(
        filename=filename,
        file=file,
        text=text,
        url=url,
        encoding=encoding,
    )
    sections = parse_elements(
        elements,
        file_type="html",
        combine_section_under_n_chars=combine_section_under_n_chars,
        new_after_n_chars=new_after_n_chars,
    )

    k_items = transform_to_k_items(sections, file_type="html")

    k_items = split_k_items(k_items, split_function=split_function)

    return k_items



from typing import List
import json

from openai import OpenAI


SYSTEM_PROMPT_QUESTIONS = """Please generate {} question asking for the key information in the user passage.
Please ask specifics questions instead of general questions, like
'What is the key information in the given paragraph?'.
The questions should be in the same language as the user passage.
Please call the process_questions function to submit the generated questions."""


TOOLS_QUESTIONS = [
    {
        "type": "function",
        "function": {
            "name": "submit_questions",
            "description": "Submits questions for the user passage.",
            "parameters": {
                "type": "object",
                "properties": {
                    "questions": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "description": "An array of questions."
                    },
                },
                "required": ["questions"]
            },
        },
    }
]


SYSTEM_PROMPT_QUESTION = """Please generate a question asking for the key information in the user passage.
Please ask a specific question instead of general questions, like
'What is the key information in the given paragraph?'.
The question should be in the same language as the user passage.
Please call the process_questions function to submit the generated question."""

TOOLS_QUESTION = [
    {
        "type": "function",
        "function": {
            "name": "submit_question",
            "description": "Submits question for the user passage.",
            "parameters": {
                "type": "object",
                "properties": {
                    "question": {
                        "type": "string",
                        "description": "The question to submit.",
                    },
                },
                "required": ["question"]
            },
        },
    }
]


models_dict = {
    'en': {
        'gpt-4': 'gpt-4-1106-preview',
        'gpt-3.5': 'gpt-3.5-turbo-1106'
    },
    'multilingual': {
        'gpt-4': 'gpt-4-0613', # the 1106 version returns bad characters https://community.openai.com/t/gpt-4-1106-preview-messes-up-function-call-parameters-encoding/478500
        # no 3.5 version before 1106 for function calls
    }
}


class QueryGenerator:
    def __init__(self, api_key, lang: str = 'en', model: str = 'gpt-4', base_url: str = None) -> None:
        """
        Class to generate questions from a given passage.
        Currently it only uses the OpenAI API.
        Parameters
        ----------
        api_key: str
            OpenAI API key.
        lang: str
            Language of the model to use.
        model: str
            Model to use.
        base_url: str
            Base url to use for the OpenAI client.
        """

        lang = lang if lang in models_dict else 'multilingual'

        assert lang in models_dict.keys(), f'No model available for {lang}'
        assert model in models_dict[lang].keys(), f'No model available for {lang} and {model}'

        self.model_name = models_dict[lang][model]
        
        if base_url is None:
            self.client = OpenAI(api_key=api_key)
        else:
            self.client = OpenAI(base_url=base_url)

    def __call__(self, passage: str, n_queries: int = 5) -> List[str]:
        """
        Generate questions from a given passage.
        Parameters
        ----------
        passage: str
            The passage to generate questions from.
        n_queries: int
            Number of questions to generate.
        Returns
        -------
        List[str]
            List of generated questions.
        """
        

        if n_queries == 1:
            system_prompt = SYSTEM_PROMPT_QUESTION
            tools = TOOLS_QUESTION
        else:
            system_prompt = SYSTEM_PROMPT_QUESTIONS.format(n_queries)
            tools = TOOLS_QUESTIONS
            tools[0]['function']['parameters']['properties']['questions']['minItems'] = n_queries
            tools[0]['function']['parameters']['properties']['questions']['maxItems'] = n_queries

        
        messages = [
            {'role': 'system', 'content': system_prompt},
            {'role': 'user', 'content': passage},
        ]

        response = self.client.chat.completions.create(
            model=self.model_name,
            messages=messages,
            tools=tools,
            tool_choice='auto'
        )

        args = json.loads(response.choices[0].message.tool_calls[0].function.arguments)

        return args['questions'] if n_queries > 1 else args['question']


from logging import getLogger
from typing import List, Dict
from chat_rag.llms import RAGLLM

logger = getLogger(__name__)

class AsyncRAG:
    """ Class for generating responses using the Retrieval-Augmented Generation (RAG) pattern in an asynchronous manner. """
    def __init__(
        self,
        retriever,
        llm_model: RAGLLM,
        lang: str = "en",
    ):
        """
        Parameters
        ----------
        retriever :
            Retriever object for retrieving contexts.
        llm_model : RAGLLM
            Language model for generating responses.
        lang : str, optional
            Language of the language model, by default "en"
        """

        self.retriever = retriever
        self.model = llm_model
        self.lang = lang

    async def retrieve(self, message: str, prev_contents: List[str], prompt_structure_dict: dict):
        """
        Retrieve new contexts if needed.

        Parameters
        ----------
        message : str
            User message.
        prev_contents : List[str]
            List of previous contexts.
        prompt_structure_dict : dict
            Dictionary containing the structure of the prompt.

        Returns
        -------
        Tuple[List[str], List[Dict[str, str]]]
            List of all conversation contexts and list of the retrieved contexts for the current user message.
        """
        logger.info("Retrieving new contexts")
        contexts = await self.retriever.retrieve(
            message,
            top_k=prompt_structure_dict["n_contexts_to_use"]
        )

        if not contexts:
            return [], []

        contents = [context["content"] for context in contexts]
        returned_contexts = [contexts[:prompt_structure_dict["n_contexts_to_use"]]]  # structure for references

        # Use a list comprehension to preserve order and not adding duplicates
        seen = set()
        contents = [x for x in prev_contents + contents if not (x in seen or seen.add(x))]
        return contents, returned_contexts

    async def stream(self, messages: List[Dict[str, str]], prev_contents: List[str], prompt_structure_dict: dict, generation_config_dict: dict, stop_words: List[str] = None):
        # Retrieve
        contents, returned_contexts = await self.retrieve(messages[-1]['content'], prev_contents, prompt_structure_dict)

        # Generate
        async for new_text in self.model.stream(
            messages, contents, prompt_structure_dict=prompt_structure_dict,
            generation_config_dict=generation_config_dict, lang=self.lang, stop_words=stop_words
        ):
            yield {"res": new_text, "context": returned_contexts}

    async def generate(self, messages: List[Dict[str, str]], prev_contents: List[str], prompt_structure_dict: dict, generation_config_dict: dict, stop_words: List[str] = None):
        # Retrieve
        contents, returned_contexts = await self.retrieve(messages[-1]['content'], prev_contents, prompt_structure_dict)

        output_text = await self.model.generate(
            messages, contents, prompt_structure_dict=prompt_structure_dict,
            generation_config_dict=generation_config_dict, lang=self.lang, stop_words=stop_words
        )
        return {"res": output_text, "context": returned_contexts}


from chat_rag.inf_retrieval.retrievers.semantic_retriever import SemanticRetriever
from chat_rag.inf_retrieval.retrievers.rerank_retriever import ReRankRetriever
from chat_rag.inf_retrieval.retrievers.retriever_client import RetrieverClient


from chat_rag.llms.base_llm import CONTEXT_PREFIX, RAGLLM
from chat_rag.llms.claude_client import AsyncClaudeChatModel, ClaudeChatModel
from chat_rag.llms.ggml_llm import GGMLModel
from chat_rag.llms.hf_llm import HFModel
from chat_rag.llms.mistral_client import AsyncMistralChatModel, MistralChatModel
from chat_rag.llms.openai_client import AsyncOpenAIChatModel, OpenAIChatModel
from chat_rag.llms.vllm_client import AsyncVLLMModel, VLLMModel
